/* ****************************************************************
   rc5dec.c 0.2.0 - Interrupt driven RC5 decoder for Atmel AVR C's.
   Website: http://markh.de/rc5dec/
   Copyright (C) 2003  Mark Haemmerling <rc5@markh.de>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Preconfigured for INT1 (PD3) and Timer/Counter0.
**************************************************************** */

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

// set your clock speed here
#define XTAL 8000000

// you may have to experiment a little with this
// in case of heavy disturbance, try values around 60
// try to use odd values (for symmetrical level detection)
#define RC5RESAMPLE 11

// this autoconfigures all timer values:
#if XTAL < 285000
# define PRESCALEDIV 1
# define TIMERCR (1<<CS00)
#elif XTAL < 2300000
# define PRESCALEDIV 8
# define TIMERCR (1<<CS01)
#elif XTAL < 18000000
# define PRESCALEDIV 64
# define TIMERCR (1<<CS01) | (1<<CS00)
#else
# define PRESCALEDIV 256
# define TIMERCR (1<<CS02)
#endif
#define RC5TIMERFIRST 0x100-((unsigned char) (XTAL/PRESCALEDIV*0.0008890))
#define RC5TIMERSECOND 0x100-((unsigned char) (XTAL/PRESCALEDIV*0.0004445))
#define RC5TIMERCANCEL 0x100-((unsigned char) (XTAL/PRESCALEDIV*0.0008890))

volatile unsigned int rc5_code = 0;
volatile unsigned char rc5_first_half;

SIGNAL (SIG_INTERRUPT1) {
/*	external interrupt handler
	edge from IR receiver detected.
	this is assumed to be the middle of a bit.
*/
	unsigned char i;
	
	if (MCUCR & (1<<ISC10))
	{
		
		for (i = RC5RESAMPLE; i; i--) if (!(PIND & (1<<PIND3))) return;
	}
	
	else
	{
		
		for (i = RC5RESAMPLE; i; i--) if ((PIND & (1<<PIND3))) return;
	}

	TCNT0 = RC5TIMERSECOND;
	rc5_first_half = 0;
	TIMSK |= (1<<TOIE0);
    GICR &= (~(1<<INT1));
}

SIGNAL (SIG_OVERFLOW0) {
/*	timer overflow handler
	sample the present level of the interrupt line.
*/	
	static unsigned int rc5_shift, rc5_temp;
	static unsigned char rc5_bit = 0, level, first_level = 1;
	unsigned char i = 0, temp;

	// resample to filter out spikes
	for (temp = RC5RESAMPLE; temp; temp--) if (PIND & (1<<PIND3)) i++;
	level = (i > ((unsigned char) (RC5RESAMPLE/2)));

	if (rc5_first_half == 2) {
		// this is a timeout, cancel operation
		goto rc5_cancel;
	} else if (rc5_first_half == 1) {
		// this sample is taken in the first half of the bit (1/4 bit time)
		first_level = level;
		if (level) MCUCR &= (~(1<<ISC10));
		else MCUCR |= (1<<ISC10);
		rc5_first_half = 2;
		GICR |= (1<<INT1);
		TCNT0 = RC5TIMERCANCEL;
	} else {
		// this sample is taken in the second half of the bit (3/4 bit time)
		TCNT0 = RC5TIMERFIRST;
		rc5_first_half = 1;
		if (first_level != level) {
			// levels differ -> valid manchester encoded bit detected
			if (!rc5_bit) {
				// this is the first bit, reset values
				rc5_temp = 0;
				rc5_shift = 0x8000;
			}
			if (!level) rc5_temp |= rc5_shift; // low level (low-active) so store a '1'
			else if (rc5_bit < 2) goto rc5_cancel; // high level ('0' bit) as startbit -> error
			if (++rc5_bit == 14) {
				// reception complete, stop operation
				rc5_code = rc5_temp;
				goto rc5_cancel;
			}
			if (rc5_shift == 0x0100) rc5_shift = 0x0020;
			else rc5_shift >>= 1;
		} else {
			// error in manchester stream -> cancel operation
			goto rc5_cancel;
		}
	}
	return;

rc5_cancel:
    MCUCR &= (~(1<<ISC10));
	GICR |= (1<<INT1);
	GIFR = (1<<INTF1);
	TIMSK &= (~(1<<TOIE0));
	first_level = 1;
	rc5_bit = 0;
}


void rc5_init(void) {
/*	init for ext. int. and timer.
	call this once to activate the rc5 decoder.
*/
	TCCR0 = TIMERCR;
	DDRD &= (~(1<<3));
	PORTD &= (~(1<<3));
	TIMSK &= (~(1<<TOIE0));
	MCUCR |= (1<<ISC11);
	MCUCR &= (~(1<<ISC10));
	GIFR = (1<<INTF1);
	GICR |= (1<<INT1);
}

int main(void) {

	rc5_init();
	
	DDRA = 0xff;
	PORTA = 0x00;
	
	DDRC = 0xff;
	PORTC = 0x00;
	
	while(1)
	{
		if( rc5_code != 0 )
		{
			// ...do some action here...
			PORTA = (unsigned char) rc5_code; // output low-byte at PORTB
			PORTC = (unsigned char) rc5_code; // output low-byte at PORTC
			rc5_code = 0; // clear code
		}
	}
}
